# -*- coding: utf-8 -*-
"""
Created on Sun Feb 28 19:06:41 2016

@name:      MultiNomial Scobit Model
@author:    Timothy Brathwaite
@summary:   Contains functions necessary for estimating multinomial scobit
            models (with the help of the "base_multinomial_cm.py" file).
            Differs from version one since it works with the shape, intercept,
            index coefficient partitioning of estimated parameters as opposed
            to the shape, index coefficient partitioning scheme of version 1.
"""
from __future__ import absolute_import

from functools import partial
import warnings
import numpy as np
from scipy.sparse import diags

from . import base_multinomial_cm_v2 as base_mcm
from .estimation import LogitTypeEstimator
from .estimation import estimate
from .display_names import model_type_to_display_name as display_name_dict

try:
    # in Python 3 range returns an iterator instead of list
    # to maintain backwards compatibility use "old" version of range
    from past.builtins import range
except ImportError:
    pass

# Define the boundary values which are not to be exceeded during computation
max_comp_value = 1e300
min_comp_value = 1e-300

max_exp = 700
min_exp = -700

# Create a warning string that will be issued if ridge regression is performed.
_msg = "NOTE: An L2-penalized regression is being performed. The "
_msg_2 = "reported standard errors and robust standard errors "
_msg_3 = "***WILL BE INCORRECT***."
_ridge_warning_msg = _msg + _msg_2 + _msg_3

# Create a warning string that will be issued if "shape_ref_pos" is passed to
# the MNSL constructor as a kwarg
_msg_4 = "The Multinomial Scobit model estimates all shape parameters"
_msg_5 = ", so shape_ref_pos will be ignored if passed."
_shape_ref_msg = _msg_4 + _msg_5


def split_param_vec(param_vec, rows_to_alts, design, return_all_types=False):
    """
    Parameters
    ----------
    param_vec : 1D ndarray.
        Elements should all be ints, floats, or longs. Should have as many
        elements as there are parameters being estimated.
    rows_to_alts : 2D scipy sparse matrix.
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset. All
        elements should be zeros or ones.
    design : 2D ndarray.
        There should be one row per observation per available alternative.
        There should be one column per utility coefficient being estimated. All
        elements should be ints, floats, or longs.
    return_all_types : bool, optional.
        Determines whether or not a tuple of 4 elements will be returned (with
        one element for the nest, shape, intercept, and index parameters for
        this model). If False, a tuple of 3 elements will be returned, as
        described below.

    Returns
    -------
    `(shapes, intercepts, betas)` : tuple of 1D ndarrays.
        The first element will be an array of the shape parameters for this
        model. The second element will either be an array of the "outside"
        intercept parameters for this model or None. The third element will be
        an array of the index coefficients for this model.

    Note
    ----
    If `return_all_types == True` then the function will return a tuple of four
    objects. In order, these objects will either be None or the arrays
    representing the arrays corresponding to the nest, shape, intercept, and
    index parameters.
    """
    # Figure out how many possible alternatives exist in the dataset
    num_shapes = rows_to_alts.shape[1]
    # Figure out how many parameters are in the index
    num_index_coefs = design.shape[1]

    # Isolate the initial shape parameters from the betas
    shapes = param_vec[:num_shapes]
    betas = param_vec[-1 * num_index_coefs:]

    # Get the remaining outside intercepts if there are any
    remaining_idx = param_vec.shape[0] - (num_shapes + num_index_coefs)
    if remaining_idx > 0:
        intercepts = param_vec[num_shapes: num_shapes + remaining_idx]
    else:
        intercepts = None

    if return_all_types:
        return None, shapes, intercepts, betas
    else:
        return shapes, intercepts, betas


def _scobit_utility_transform(systematic_utilities,
                              alt_IDs,
                              rows_to_alts,
                              shape_params,
                              intercept_params,
                              intercept_ref_pos=None,
                              *args, **kwargs):
    """
    Parameters
    ----------
    systematic_utilities : 1D ndarray.
        All elements should be ints, floats, or longs. Should contain the
        systematic utilities of each observation per available alternative.
        Note that this vector is formed by the dot product of the design matrix
        with the vector of utility coefficients.
    alt_IDs : 1D ndarray.
        All elements should be ints. There should be one row per obervation per
        available alternative for the given observation. Elements denote the
        alternative corresponding to the given row of the design matrix.
    rows_to_alts : 2D scipy sparse matrix.
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset. All
        elements should be zeros or ones.
    shape_params : None or 1D ndarray.
        If an array, each element should be an int, float, or long. There
        should be one value per shape parameter of the model being used.
    intercept_params : None or 1D ndarray.
        If an array, each element should be an int, float, or long. If J is the
        total number of possible alternatives for the dataset being modeled,
        there should be J-1 elements in the array.
    intercept_ref_pos : int, or None, optional.
        Specifies the index of the alternative, in the ordered array of unique
        alternatives, that is not having its intercept parameter estimated (in
        order to ensure identifiability). Should only be None if
        `intercept_params` is None.

    Returns
    -------
    transformations : 2D ndarray.
        Should have shape `(systematic_utilities.shape[0], 1)`. The returned
        array contains the transformed utility values for this model. All
        elements should be ints, floats, or longs.
    """
    # Figure out what indices are to be filled in
    if intercept_ref_pos is not None and intercept_params is not None:
        needed_idxs = range(intercept_params.shape[0] + 1)
        needed_idxs.remove(intercept_ref_pos)

        if len(intercept_params.shape) > 1 and intercept_params.shape[1] > 1:
            # Get an array of zeros with shape
            # (num_possible_alternatives, num_parameter_samples)
            all_intercepts = np.zeros((rows_to_alts.shape[1],
                                       intercept_params.shape[1]))
            # For alternatives having their intercept estimated, replace the
            # zeros with the current value of the estimated intercepts
            all_intercepts[needed_idxs, :] = intercept_params
        else:
            # Get an array of zeros with shape (num_possible_alternatives,)
            all_intercepts = np.zeros(rows_to_alts.shape[1])
            # For alternatives having their intercept estimated, replace the
            # zeros with the current value of the estimated intercepts
            all_intercepts[needed_idxs] = intercept_params
    else:
        # Create a full set of intercept parameters including the intercept
        # constrained to zero
        all_intercepts = np.zeros(rows_to_alts.shape[1])

    # Figure out what intercept values correspond to each row of the
    # systematic utilities
    long_intercepts = rows_to_alts.dot(all_intercepts)

    # Convert the shape parameters back into their 'natural parametrization'
    natural_shapes = np.exp(shape_params)
    natural_shapes[np.isposinf(natural_shapes)] = max_comp_value
    # Figure out what shape values correspond to each row of the
    # systematic utilities
    long_natural_shapes = rows_to_alts.dot(natural_shapes)

    # Calculate the data dependent part of the transformation
    # Also, along the way, guard against numeric underflow or overflow
    exp_neg_v = np.exp(-1 * systematic_utilities)
    exp_neg_v[np.isposinf(exp_neg_v)] = max_comp_value

    powered_term = np.power(1 + exp_neg_v, long_natural_shapes)
    powered_term[np.isposinf(powered_term)] = max_comp_value

    term_2 = np.log(powered_term - 1)
    # Guard against overvlow
    too_big_idx = np.isposinf(powered_term)
    term_2[too_big_idx] = (-1 * long_natural_shapes[too_big_idx] *
                           systematic_utilities[too_big_idx])

    transformations = long_intercepts - term_2
    # Guard against overflow
    transformations[np.isposinf(transformations)] = max_comp_value
    transformations[np.isneginf(transformations)] = -1 * max_comp_value

    # Be sure to return a 2D array since other functions will be expecting that
    if len(transformations.shape) == 1:
        transformations = transformations[:, np.newaxis]

    return transformations


def _scobit_transform_deriv_v(systematic_utilities,
                              alt_IDs,
                              rows_to_alts,
                              shape_params,
                              output_array=None,
                              *args, **kwargs):
    """
    Parameters
    ----------
    systematic_utilities : 1D ndarray.
        All elements should be ints, floats, or longs. Should contain the
        systematic utilities of each observation per available alternative.
        Note that this vector is formed by the dot product of the design matrix
        with the vector of utility coefficients.
    alt_IDs : 1D ndarray.
        All elements should be ints. There should be one row per obervation per
        available alternative for the given observation. Elements denote the
        alternative corresponding to the given row of the design matrix.
    rows_to_alts : 2D scipy sparse matrix.
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset. All
        elements should be zeros or ones.
    shape_params : None or 1D ndarray.
        If an array, each element should be an int, float, or long. There
        should be one value per shape parameter of the model being used.
    output_array : 2D scipy sparse array.
        The array should be square and it should have
        `systematic_utilities.shape[0]` rows. It's data is to be replaced with
        the correct derivatives of the transformation vector with respect to
        the vector of systematic utilities. This argument is NOT optional.

    Returns
    -------
    output_array : 2D scipy sparse array.
        The shape of the returned array is `(systematic_utilities.shape[0],
        systematic_utilities.shape[0])`. The returned array specifies the
        derivative of the transformed utilities with respect to the systematic
        utilities. All elements are ints, floats, or longs.
    """
    # Note  the np.exp is  needed because the raw curvature params are the log
    # of the 'natural' curvature params. This is to ensure the natural shape
    # params are always positive
    curve_shapes = np.exp(shape_params)
    curve_shapes[np.isposinf(curve_shapes)] = max_comp_value
    long_curve_shapes = rows_to_alts.dot(curve_shapes)

    # Generate the needed terms for the derivative of the transformation with
    # respect to the systematic utility and guard against underflow or overflow
    exp_neg_v = np.exp(-1 * systematic_utilities)

    powered_term = np.power(1 + exp_neg_v, long_curve_shapes)

    small_powered_term = np.power(1 + exp_neg_v, long_curve_shapes - 1)

    derivs = (long_curve_shapes *
              exp_neg_v *
              small_powered_term /
              (powered_term - 1))
    # Use L'Hopitals rule to deal with overflow from v --> -inf
    # From plots, the assignment below may also correctly handle cases where we
    # have overflow from moderate v (say |v| <= 10) and large shape parameters.
    too_big_idx = (np.isposinf(derivs) +
                   np.isposinf(exp_neg_v) +
                   np.isposinf(powered_term) +
                   np.isposinf(small_powered_term)).astype(bool)
    derivs[too_big_idx] = long_curve_shapes[too_big_idx]
    # Use L'Hopitals rule to deal with underflow from v --> inf
    too_small_idx = np.where((exp_neg_v == 0) | (powered_term - 1 == 0))
    derivs[too_small_idx] = 1.0

    # Assign the calculated derivatives to the output array
    output_array.data = derivs
    assert output_array.shape == (systematic_utilities.shape[0],
                                  systematic_utilities.shape[0])

    # Return the matrix of dh_dv. Note the off-diagonal entries are zero
    # because each transformation only depends on its value of v and no others
    return output_array


def _scobit_transform_deriv_shape(systematic_utilities,
                                  alt_IDs,
                                  rows_to_alts,
                                  shape_params,
                                  output_array=None,
                                  *args, **kwargs):
    """
    Parameters
    ----------
    systematic_utilities : 1D ndarray.
        All elements should be ints, floats, or longs. Should contain the
        systematic utilities of each observation per available alternative.
        Note that this vector is formed by the dot product of the design matrix
        with the vector of utility coefficients.
    alt_IDs : 1D ndarray.
        All elements should be ints. There should be one row per obervation per
        available alternative for the given observation. Elements denote the
        alternative corresponding to the given row of the design matrix.
    rows_to_alts : 2D scipy sparse matrix.
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset. All
        elements should be zeros or ones.
    shape_params : None or 1D ndarray.
        If an array, each element should be an int, float, or long. There
        should be one value per shape parameter of the model being used.
    output_array : 2D scipy sparse array.
        The array should have shape `(systematic_utilities.shape[0],
        shape_params.shape[0])`. It's data is to be replaced with the correct
        derivatives of the transformation vector with respect to the vector of
        shape parameters. This argument is NOT optional.

    Returns
    -------
    output_array : 2D scipy sparse array.
        The shape of the returned array is `(systematic_utilities.shape[0],
        shape_params.shape[0])`. The returned array specifies the derivative of
        the transformed utilities with respect to the shape parameters. All
        elements are ints, floats, or longs.
    """
    # Note the np.exp is needed because the raw curvature params are the log
    # of the 'natural' curvature params. This is to ensure the natural shape
    # params are always positive.
    curve_shapes = np.exp(shape_params)
    curve_shapes[np.isposinf(curve_shapes)] = max_comp_value
    long_curve_shapes = rows_to_alts.dot(curve_shapes)

    # Generate the needed terms for the derivative of the transformation with
    # respect to the systematic utility and guard against underflow or overflow
    exp_neg_v = np.exp(-1 * systematic_utilities)

    powered_term = np.power(1 + exp_neg_v, long_curve_shapes)

    # Note the "* long_curve_shapes" accounts for the fact that the derivative
    # of the natural shape params (i.e. exp(shape_param)) with respect to the
    # shape param is simply exp(shape_param) = natural shape param. The
    # multipication comes from the chain rule.
    curve_derivs = (-1 * np.log1p(exp_neg_v) *
                    powered_term / (powered_term - 1)) * long_curve_shapes
    # Note that all of the following overflow and underflow guards are based
    # on L'Hopital's rule.
    # Guard against underflow from v --> inf. Note we end up with '-1' after
    # accounting for the jacobian.
    too_big_idx = np.where((powered_term - 1) == 0)
    curve_derivs[too_big_idx] = -1
    # Guard against overflow from v --> -inf.
    too_small_idx = np.isposinf(exp_neg_v)
    curve_derivs[too_small_idx] = max_comp_value
    # Guard against overflow from moderate v but shape --> inf.
    shape_too_big_idx = np.where((np.abs(systematic_utilities) <= 10) &
                                 np.isposinf(powered_term))
    curve_derivs[shape_too_big_idx] =\
        (-1 * np.log1p(exp_neg_v))[shape_too_big_idx]

    # Assign the calculated derivatives to the output array
    output_array.data = curve_derivs

    return output_array


def _scobit_transform_deriv_alpha(systematic_utilities,
                                  alt_IDs,
                                  rows_to_alts,
                                  intercept_params,
                                  output_array=None,
                                  *args, **kwargs):
    """
    Parameters
    ----------
    systematic_utilities : 1D ndarray.
        All elements should be ints, floats, or longs. Should contain the
        systematic utilities of each observation per available alternative.
        Note that this vector is formed by the dot product of the design matrix
        with the vector of utility coefficients.
    alt_IDs : 1D ndarray.
        All elements should be ints. There should be one row per obervation per
        available alternative for the given observation. Elements denote the
        alternative corresponding to the given row of the design matrix.
    rows_to_alts : 2D scipy sparse matrix.
        There should be one row per observation per available alternative and
        one column per possible alternative. This matrix maps the rows of the
        design matrix to the possible alternatives for this dataset. All
        elements should be zeros or ones.
    intercept_params : 1D ndarray or None.
        If an array, each element should be an int, float, or long. For
        identifiability, there should be J- 1 elements where J is the total
        number of observed alternatives for this dataset.
    output_array: None or 2D scipy sparse array.
        If a sparse array is pased, it should contain the derivative of the
        vector of transformed utilities with respect to the intercept
        parameters outside of the index. This keyword argurment will be
        returned. If there are no intercept parameters outside of the index,
        then `output_array` should equal None. If there are intercept
        parameters outside of the index, then `output_array` should be
        `rows_to_alts` with the all of its columns except the column
        corresponding to the alternative whose intercept is not being estimated
        in order to ensure identifiability.

    Returns
    -------
    output_array.
    """
    return output_array


def create_calc_dh_dv(estimator):
    """
    Return the function that can be used in the various gradient and hessian
    calculations to calculate the derivative of the transformation with respect
    to the index.

    Parameters
    ----------
    estimator : an instance of the estimation.LogitTypeEstimator class.
        Should contain a `design` attribute that is a 2D ndarray representing
        the design matrix for this model and dataset.

    Returns
    -------
    Callable.
        Will accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, (shape parameters if there are any) and miscellaneous
        args and kwargs. Should return a 2D array whose elements contain the
        derivative of the tranformed utility vector with respect to the vector
        of systematic utilities. The dimensions of the returned vector should
        be `(design.shape[0], design.shape[0])`.
    """
    dh_dv = diags(np.ones(estimator.design.shape[0]), 0, format='csr')
    # Create a function that will take in the pre-formed matrix, replace its
    # data in-place with the new data, and return the correct dh_dv on each
    # iteration of the minimizer
    calc_dh_dv = partial(_scobit_transform_deriv_v, output_array=dh_dv)
    return calc_dh_dv


def create_calc_dh_d_shape(estimator):
    """
    Return the function that can be used in the various gradient and hessian
    calculations to calculate the derivative of the transformation with respect
    to the shape parameters.

    Parameters
    ----------
    estimator : an instance of the estimation.LogitTypeEstimator class.
        Should contain a `rows_to_alts` attribute that is a 2D scipy sparse
        matrix that maps the rows of the `design` matrix to the alternatives
        available in this dataset.

    Returns
    -------
    Callable.
        Will accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, (shape parameters if there are any) and miscellaneous
        args and kwargs. Should return a 2D array whose elements contain the
        derivative of the tranformed utility vector with respect to the vector
        of shape parameters. The dimensions of the returned vector should
        be `(design.shape[0], num_alternatives)`.
    """
    dh_d_shape = estimator.rows_to_alts.copy()
    # Create a function that will take in the pre-formed matrix, replace its
    # data in-place with the new data, and return the correct dh_dshape on each
    # iteration of the minimizer
    calc_dh_d_shape = partial(_scobit_transform_deriv_shape,
                              output_array=dh_d_shape)
    return calc_dh_d_shape


def create_calc_dh_d_alpha(estimator):
    """
    Return the function that can be used in the various gradient and hessian
    calculations to calculate the derivative of the transformation with respect
    to the outside intercept parameters.

    Parameters
    ----------
    estimator : an instance of the estimation.LogitTypeEstimator class.
        Should contain a `rows_to_alts` attribute that is a 2D scipy sparse
        matrix that maps the rows of the `design` matrix to the alternatives
        available in this dataset. Should also contain an `intercept_ref_pos`
        attribute that is either None or an int. This attribute should denote
        which intercept is not being estimated (in the case of outside
        intercept parameters) for identification purposes.

    Returns
    -------
    Callable.
        Will accept a 1D array of systematic utility values, a 1D array of
        alternative IDs, (shape parameters if there are any) and miscellaneous
        args and kwargs. Should return a 2D array whose elements contain the
        derivative of the tranformed utility vector with respect to the vector
        of outside intercepts. The dimensions of the returned vector should
        be `(design.shape[0], num_alternatives - 1)`.
    """
    if estimator.intercept_ref_pos is not None:
        needed_idxs = range(estimator.rows_to_alts.shape[1])
        needed_idxs.remove(estimator.intercept_ref_pos)
        dh_d_alpha = (estimator.rows_to_alts
                               .copy()
                               .transpose()[needed_idxs, :]
                               .transpose())
    else:
        dh_d_alpha = None
    # Create a function that will take in the pre-formed matrix, replace its
    # data in-place with the new data, and return the correct dh_dalpha on each
    # iteration of the minimizer
    calc_dh_d_alpha = partial(_scobit_transform_deriv_alpha,
                              output_array=dh_d_alpha)

    return calc_dh_d_alpha


class ScobitEstimator(LogitTypeEstimator):
    """
    Estimation Object used to enforce uniformity in the estimation process
    across the various logit-type models.

    Parameters
    ----------
    model_obj : a pylogit.base_multinomial_cm_v2.MNDC_Model instance.
        Should contain the following attributes:

          - alt_IDs
          - choices
          - design
          - intercept_ref_position
          - shape_ref_position
          - utility_transform
    mapping_res : dict.
        Should contain the scipy sparse matrices that map the rows of the long
        format dataframe to various other objects such as the available
        alternatives, the unique observations, etc. The keys that it must have
        are `['rows_to_obs', 'rows_to_alts', 'chosen_row_to_obs']`
    ridge : int, float, long, or None.
            Determines whether or not ridge regression is performed. If a
            scalar is passed, then that scalar determines the ridge penalty for
            the optimization. The scalar should be greater than or equal to
            zero..
    zero_vector : 1D ndarray.
        Determines what is viewed as a "null" set of parameters. It is
        explicitly passed because some parameters (e.g. parameters that must be
        greater than zero) have their null values at values other than zero.
    split_params : callable.
        Should take a vector of parameters, `mapping_res['rows_to_alts']`, and
        model_obj.design as arguments. Should return a tuple containing
        separate arrays for the model's shape, outside intercept, and index
        coefficients. For each of these arrays, if this model does not contain
        the particular type of parameter, the callable should place a `None` in
        its place in the tuple.
    constrained_pos : list or None, optional.
        Denotes the positions of the array of estimated parameters that are
        not to change from their initial values. If a list is passed, the
        elements are to be integers where no such integer is greater than
        `num_params` Default == None.
    weights : 1D ndarray or None, optional.
        Allows for the calculation of weighted log-likelihoods. The weights can
        represent various things. In stratified samples, the weights may be
        the proportion of the observations in a given strata for a sample in
        relation to the proportion of observations in that strata in the
        population. In latent class models, the weights may be the probability
        of being a particular class.
    """
    def set_derivatives(self):
        self.calc_dh_dv = create_calc_dh_dv(self)
        self.calc_dh_d_alpha = create_calc_dh_d_alpha(self)
        self.calc_dh_d_shape = create_calc_dh_d_shape(self)

    def check_length_of_initial_values(self, init_values):
        """
        Ensures that `init_values` is of the correct length. Raises a helpful
        ValueError if otherwise.

        Parameters
        ----------
        init_values : 1D ndarray.
            The initial values to start the optimization process with. There
            should be one value for each index coefficient, outside intercept
            parameter, and shape parameter being estimated.

        Returns
        -------
        None.
        """
        # Calculate the expected number of shape and index parameters
        # Note the scobit model has one shape parameter per alternative.
        num_alts = self.rows_to_alts.shape[1]
        num_index_coefs = self.design.shape[1]

        if self.intercept_ref_pos is not None:
            assumed_param_dimensions = num_index_coefs + 2 * num_alts - 1
        else:
            assumed_param_dimensions = num_index_coefs + num_alts

        if init_values.shape[0] != assumed_param_dimensions:
            msg_1 = "The initial values are of the wrong dimension."
            msg_2 = "It should be of dimension {}"
            msg_3 = "But instead it has dimension {}"
            raise ValueError(msg_1 +
                             msg_2.format(assumed_param_dimensions) +
                             msg_3.format(init_values.shape[0]))

        return None


class MNSL(base_mcm.MNDC_Model):
    """
    Parameters
    ----------
    data : string or pandas dataframe.
        If string, data should be an absolute or relative path to a CSV file
        containing the long format data for this choice model. Note long format
        is has one row per available alternative for each observation. If
        pandas dataframe, the dataframe should be the long format data for the
        choice model.
    alt_id_col :str.
        Should denote the column in data which contains the alternative
        identifiers for each row.
    obs_id_col : str.
        Should denote the column in data which contains the observation
        identifiers for each row.
    choice_col : str.
        Should denote the column in data which contains the ones and zeros that
        denote whether or not the given row corresponds to the chosen
        alternative for the given individual.
    specification : OrderedDict.
        Keys are a proper subset of the columns in `data`. Values are either a
        list or a single string, "all_diff" or "all_same". If a list, the
        elements should be:
            - single objects that are in the alternative ID column of `data`
            - lists of objects that are within the alternative ID column of
              `data`. For each single object in the list, a unique column will
              be created (i.e. there will be a unique coefficient for that
              variable in the corresponding utility equation of the
              corresponding alternative). For lists within the
              `specification` values, a single column will be created for all
              the alternatives within the iterable (i.e. there will be one
              common coefficient for the variables in the iterable).
    intercept_ref_pos : int, optional.
         Valid only when the intercepts being estimated are not part of the
         index. Specifies the alternative in the ordered array of unique
         alternative ids whose intercept or alternative-specific constant is
         not estimated, to ensure model identifiability. Default == None.
    names : OrderedDict, optional.
        Should have the same keys as `specification`. For each key:
            - if the corresponding value in `specification` is "all_same", then
              there should be a single string as the value in names.
            - if the corresponding value in `specification` is "all_diff", then
              there should be a list of strings as the value in names. There
              should be one string in the value in names for each possible
              alternative.
            - if the corresponding value in `specification` is a list, then
              there should be a list of strings as the value in names. There
              should be one string the value in names per item in the value in
              `specification`.
        Default == None.
    intercept_names : list, or None, optional.
        If a list is passed, then the list should have the same number of
        elements as there are possible alternatives in data, minus 1. Each
        element of the list should be a string--the name of the corresponding
        alternative's intercept term, in sorted order of the possible
        alternative IDs. If None is passed, the resulting names that are shown
        in the estimation results will be
        `["Outside_ASC_{}".format(x) for x in shape_names]`. Default = None.
    shape_names : list, or None, optional.
        If a list is passed, then the list should have the same number of
        elements as there are possible alternative IDs in data. Each element of
        the list should be a string denoting the name of the corresponding
        shape parameter for the given alternative, in sorted order of the
        possible alternative IDs. The resulting names which are shown in the
        estimation results will be ["shape_{}".format(x) for x in shape_names].
        Default == None.
    """
    def __init__(self,
                 data,
                 alt_id_col,
                 obs_id_col,
                 choice_col,
                 specification,
                 intercept_ref_pos=None,
                 names=None,
                 intercept_names=None,
                 shape_names=None,
                 **kwargs):
        ##########
        # Print a helpful message if shape_ref_pos has been included
        ##########
        if "shape_ref_pos" in kwargs and kwargs["shape_ref_pos"] is not None:
            warnings.warn(_shape_ref_msg)

        ##########
        # Carry out the common instantiation process for all choice models
        ##########
        super(MNSL, self).__init__(data,
                                   alt_id_col,
                                   obs_id_col,
                                   choice_col,
                                   specification,
                                   intercept_ref_pos=intercept_ref_pos,
                                   names=names,
                                   intercept_names=intercept_names,
                                   shape_names=shape_names,
                                   model_type=display_name_dict["Scobit"])

        ##########
        # Store the utility transform function
        ##########
        self.utility_transform = partial(_scobit_utility_transform,
                                         intercept_ref_pos=intercept_ref_pos)

        return None

    def fit_mle(self,
                init_vals,
                init_shapes=None,
                init_intercepts=None,
                init_coefs=None,
                print_res=True,
                method="BFGS",
                loss_tol=1e-06,
                gradient_tol=1e-06,
                maxiter=1000,
                ridge=None,
                constrained_pos=None,
                just_point=False,
                **kwargs):
        """
        Parameters
        ----------
        init_vals : 1D ndarray.
            The initial values to start the optimization process with. There
            should be one value for each index coefficient and shape
            parameter being estimated. Shape parameters should come before
            intercept parameters, which should come before index coefficients.
            One can also pass None, and instead pass `init_shapes`, optionally
            `init_intercepts` if `"intercept"` is not in the utility
            specification, and `init_coefs`.
        init_shapes : 1D ndarray or None, optional.
            The initial values of the shape parameters. All elements should be
            ints, floats, or longs. There should be one parameter per possible
            alternative id in the dataset. This keyword argument will be
            ignored if `init_vals` is not None. Default == None.
        init_intercepts : 1D ndarray or None, optional.
            The initial values of the intercept parameters. There should be one
            parameter per possible alternative id in the dataset, minus one.
            The passed values for this argument will be ignored if `init_vals`
            is not None. This keyword argument should only be used if
            `"intercept"` is not in the utility specification. Default == None.
        init_coefs : 1D ndarray or None, optional.
            The initial values of the index coefficients. There should be one
            coefficient per index variable. The passed values for this argument
            will be ignored if `init_vals` is not None. Default == None.
        print_res : bool, optional.
            Determines whether the timing and initial and final log likelihood
            results will be printed as they they are determined.
            Default `== True`.
        method : str, optional.
            Should be a valid string for scipy.optimize.minimize. Determines
            the optimization algorithm that is used for this problem.
            Default `== 'bfgs'`.
        loss_tol : float, optional.
            Determines the tolerance on the difference in objective function
            values from one iteration to the next that is needed to determine
            convergence. Default `== 1e-06`.
        gradient_tol : float, optional.
            Determines the tolerance on the difference in gradient values from
            one iteration to the next which is needed to determine convergence.
            Default `== 1e-06`.
        maxiter : int, optional.
            Determines the maximum number of iterations used by the optimizer.
            Default `== 1000`.
        ridge : int, float, long, or None, optional.
            Determines whether or not ridge regression is performed. If a
            scalar is passed, then that scalar determines the ridge penalty for
            the optimization. The scalar should be greater than or equal to
            zero. Default `== None`.
        constrained_pos : list or None, optional.
            Denotes the positions of the array of estimated parameters that are
            not to change from their initial values. If a list is passed, the
            elements are to be integers where no such integer is greater than
            `init_vals.size.` Default == None.
        just_point : bool, optional.
            Determines whether (True) or not (False) calculations that are non-
            critical for obtaining the maximum likelihood point estimate will
            be performed. If True, this function will return the results
            dictionary from scipy.optimize. Default == False.

        Returns
        -------
        None. Estimation results are saved to the model instance.
        """
        # Store the optimization method
        self.optimization_method = method

        # Store the ridge parameter
        self.ridge_param = ridge
        if ridge is not None:
            warnings.warn(_ridge_warning_msg)

        # Construct the mappings from alternatives to observations and from
        # chosen alternatives to observations
        mapping_res = self.get_mappings_for_fit()
        rows_to_alts = mapping_res["rows_to_alts"]

        # Create init_vals from init_coefs, init_intercepts, and init_shapes if
        # those arguments are passed to the function and init_vals is None.
        if init_vals is None and all([x is not None for x in [init_shapes,
                                                              init_coefs]]):
            ##########
            # Check the integrity of the parameter kwargs
            ##########
            num_alternatives = rows_to_alts.shape[1]
            try:
                assert init_shapes.shape[0] == num_alternatives
            except AssertionError:
                msg = "init_shapes is of length {} but should be of length {}"
                raise ValueError(msg.format(init_shapes.shape,
                                            num_alternatives))

            try:
                assert init_coefs.shape[0] == self.design.shape[1]
            except AssertionError:
                msg = "init_coefs has length {} but should have length {}"
                raise ValueError(msg.format(init_coefs.shape,
                                            self.design.shape[1]))

            try:
                if init_intercepts is not None:
                    assert init_intercepts.shape[0] == (num_alternatives - 1)
            except AssertionError:
                msg = "init_intercepts has length {} but should have length {}"
                raise ValueError(msg.format(init_intercepts.shape,
                                            num_alternatives - 1))

            # The code block below will limit users to only having 'inside'
            # OR 'outside' intercept parameters but not both.
            # try:
            #     condition_1 = "intercept" not in self.specification
            #     condition_2 = init_intercepts is None
            #     assert condition_1 or condition_2
            # except AssertionError as e:
            #     msg = "init_intercepts should only be used if 'intercept' is"
            #     msg_2 = " not in one's index specification."
            #     msg_3 = "Either make init_intercepts = None or remove "
            #     msg_4 = "'intercept' from the specification."
            #     print(msg + msg_2 )
            #     print(msg_3 + msg_4)
            #     raise e

            if init_intercepts is not None:
                init_vals = np.concatenate((init_shapes,
                                            init_intercepts,
                                            init_coefs), axis=0)
            else:
                init_vals = np.concatenate((init_shapes,
                                            init_coefs), axis=0)
        elif init_vals is None:
            msg = "If init_vals is None, then users must pass both init_coefs "
            msg_2 = "and init_shapes."
            raise ValueError(msg + msg_2)

        # Create the estimation object
        zero_vector = np.zeros(init_vals.shape)
        scobit_estimator = ScobitEstimator(self,
                                           mapping_res,
                                           ridge,
                                           zero_vector,
                                           split_param_vec,
                                           constrained_pos=constrained_pos)
        # Set the derivative functions for estimation
        scobit_estimator.set_derivatives()

        # Perform one final check on the length of the initial values
        scobit_estimator.check_length_of_initial_values(init_vals)

        # Get the estimation results
        estimation_res = estimate(init_vals,
                                  scobit_estimator,
                                  method,
                                  loss_tol,
                                  gradient_tol,
                                  maxiter,
                                  print_res,
                                  just_point=just_point)

        if not just_point:
            # Store the estimation results
            self.store_fit_results(estimation_res)

            return None
        else:
            return estimation_res



"""
@author:    Timothy Brathwaite
@name:      Bootstrap Estimation Procedures
@summary:   This module provides functions that will perform the MLE for each
            of the bootstrap samples.
"""
import numpy as np
import pandas as pd

from . import pylogit as pl
from .display_names import model_type_to_display_name


def extract_default_init_vals(orig_model_obj, mnl_point_series, num_params):
    """
    Get the default initial values for the desired model type, based on the
    point estimate of the MNL model that is 'closest' to the desired model.

    Parameters
    ----------
    orig_model_obj : an instance or sublcass of the MNDC class.
        Should correspond to the actual model that we want to bootstrap.
    mnl_point_series : pandas Series.
        Should denote the point estimate from the MNL model that is 'closest'
        to the desired model.
    num_params : int.
        Should denote the number of parameters being estimated (including any
        parameters that are being constrained during estimation).

    Returns
    -------
    init_vals : 1D ndarray of initial values for the MLE of the desired model.
    """
    # Initialize the initial values
    init_vals = np.zeros(num_params, dtype=float)
    # Figure out which values in mnl_point_series are the index coefficients
    no_outside_intercepts = orig_model_obj.intercept_names is None
    if no_outside_intercepts:
        init_index_coefs = mnl_point_series.values
        init_intercepts = None
    else:
        init_index_coefs =\
            mnl_point_series.loc[orig_model_obj.ind_var_names].values
        init_intercepts =\
            mnl_point_series.loc[orig_model_obj.intercept_names].values

    # Add any mixing variables to the index coefficients.
    if orig_model_obj.mixing_vars is not None:
        num_mixing_vars = len(orig_model_obj.mixing_vars)
        init_index_coefs = np.concatenate([init_index_coefs,
                                           np.zeros(num_mixing_vars)],
                                          axis=0)

    # Account for the special transformation of the index coefficients that is
    # needed for the asymmetric logit model.
    if orig_model_obj.model_type == model_type_to_display_name["Asym"]:
        multiplier = np.log(len(np.unique(orig_model_obj.alt_IDs)))
        # Cast the initial index coefficients to a float dtype to ensure
        # successful broadcasting
        init_index_coefs = init_index_coefs.astype(float)
        # Adjust the scale of the index coefficients for the asymmetric logit.
        init_index_coefs /= multiplier

    # Combine the initial interept values with the initial index coefficients
    if init_intercepts is not None:
        init_index_coefs =\
            np.concatenate([init_intercepts, init_index_coefs], axis=0)

    # Add index coefficients (and mixing variables) to the total initial array
    num_index = init_index_coefs.shape[0]
    init_vals[-1 * num_index:] = init_index_coefs

    # Note that the initial values for the transformed nest coefficients and
    # the shape parameters is zero so we don't have to change anything
    return init_vals


def get_model_abbrev(model_obj):
    """
    Extract the string used to specify the model type of this model object in
    `pylogit.create_chohice_model`.

    Parameters
    ----------
    model_obj : An MNDC_Model instance.

    Returns
    -------
    str. The internal abbreviation used for the particular type of MNDC_Model.
    """
    # Get the 'display name' for our model.
    model_type = model_obj.model_type
    # Find the model abbreviation for this model's display name.
    for key in model_type_to_display_name:
        if model_type_to_display_name[key] == model_type:
            return key
    # If none of the strings in model_type_to_display_name matches our model
    # object, then raise an error.
    msg = "Model object has an unknown or incorrect model type."
    raise ValueError(msg)


def get_model_creation_kwargs(model_obj):
    """
    Get a dictionary of the keyword arguments needed to create the passed model
    object using `pylogit.create_choice_model`.

    Parameters
    ----------
    model_obj : An MNDC_Model instance.

    Returns
    -------
    model_kwargs : dict.
        Contains the keyword arguments and the required values that are needed
        to initialize a replica of `model_obj`.
    """
    # Extract the model abbreviation for this model
    model_abbrev = get_model_abbrev(model_obj)

    # Create a dictionary to store the keyword arguments needed to Initialize
    # the new model object.d
    model_kwargs = {"model_type": model_abbrev,
                    "names": model_obj.name_spec,
                    "intercept_names": model_obj.intercept_names,
                    "intercept_ref_pos": model_obj.intercept_ref_position,
                    "shape_names": model_obj.shape_names,
                    "shape_ref_pos": model_obj.shape_ref_position,
                    "nest_spec": model_obj.nest_spec,
                    "mixing_vars": model_obj.mixing_vars,
                    "mixing_id_col": model_obj.mixing_id_col}

    return model_kwargs


def get_mnl_point_est(orig_model_obj,
                      new_df,
                      boot_id_col,
                      num_params,
                      mnl_spec,
                      mnl_names,
                      mnl_init_vals,
                      mnl_fit_kwargs):
    """
    Calculates the MLE for the desired MNL model.

    Parameters
    ----------
    orig_model_obj : An MNDC_Model instance.
        The object corresponding to the desired model being bootstrapped.
    new_df : pandas DataFrame.
        The pandas dataframe containing the data to be used to estimate the
        MLE of the MNL model for the current bootstrap sample.
    boot_id_col : str.
        Denotes the new column that specifies the bootstrap observation ids for
        choice model estimation.
    num_params : non-negative int.
        The number of parameters in the MLE of the `orig_model_obj`.
    mnl_spec : OrderedDict or None.
        If `orig_model_obj` is not a MNL model, then `mnl_spec` should be an
        OrderedDict that contains the specification dictionary used to estimate
        the MNL model that will provide starting values for the final estimated
        model. If `orig_model_obj` is a MNL model, then `mnl_spec` may be None.
    mnl_names : OrderedDict or None.
        If `orig_model_obj` is not a MNL model, then `mnl_names` should be an
        OrderedDict that contains the name dictionary used to initialize the
        MNL model that will provide starting values for the final estimated
        model. If `orig_model_obj` is a MNL, then `mnl_names` may be None.
    mnl_init_vals : 1D ndarray or None.
        If `orig_model_obj` is not a MNL model, then `mnl_init_vals` should be
        a 1D ndarray. `mnl_init_vals` should denote the initial values used to
        estimate the MNL model that provides starting values for the final
        desired model. If `orig_model_obj` is a MNL model, then `mnl_init_vals`
        may be None.
    mnl_fit_kwargs : dict or None.
        If `orig_model_obj` is not a MNL model, then `mnl_fit_kwargs` should be
        a dict. `mnl_fit_kwargs` should denote the keyword arguments used when
        calling the `fit_mle` function of the MNL model that will provide
        starting values to the desired choice model. If `orig_model_obj` is a
        MNL model, then `mnl_fit_kwargs` may be None.

    Returns
    -------
    mnl_point : dict.
        The dictionary returned by `scipy.optimize` after estimating the
        desired MNL model.
    mnl_obj : An MNL model instance.
        The model object used to estimate the desired MNL model.
    """
    # Get specification and name dictionaries for the mnl model, for the case
    # where the model being bootstrapped is an MNL model. In this case, the
    # the mnl_spec and the mnl_names that are passed to the function are
    # expected to be None.
    if orig_model_obj.model_type == model_type_to_display_name["MNL"]:
        mnl_spec = orig_model_obj.specification
        mnl_names = orig_model_obj.name_spec
        if mnl_init_vals is None:
            mnl_init_vals = np.zeros(num_params)
        if mnl_fit_kwargs is None:
            mnl_fit_kwargs = {}

    # Alter the mnl_fit_kwargs to ensure that we only perform point estimation
    mnl_fit_kwargs["just_point"] = True
    # Use BFGS by default to estimate the MNL since it works well for the MNL.
    if "method" not in mnl_fit_kwargs:
        mnl_fit_kwargs["method"] = "BFGS"

    # Initialize the mnl model object for the given bootstrap sample.
    mnl_obj = pl.create_choice_model(data=new_df,
                                     alt_id_col=orig_model_obj.alt_id_col,
                                     obs_id_col=boot_id_col,
                                     choice_col=orig_model_obj.choice_col,
                                     specification=mnl_spec,
                                     model_type="MNL",
                                     names=mnl_names)

    # Get the MNL point estimate for the parameters of this bootstrap sample.
    mnl_point = mnl_obj.fit_mle(mnl_init_vals, **mnl_fit_kwargs)
    return mnl_point, mnl_obj


def retrieve_point_est(orig_model_obj,
                       new_df,
                       new_id_col,
                       num_params,
                       mnl_spec,
                       mnl_names,
                       mnl_init_vals,
                       mnl_fit_kwargs,
                       extract_init_vals=None,
                       **fit_kwargs):
    """
    Calculates the MLE for the desired MNL model.

    Parameters
    ----------
    orig_model_obj : An MNDC_Model instance.
        The object corresponding to the desired model being bootstrapped.
    new_df : pandas DataFrame.
        The pandas dataframe containing the data to be used to estimate the
        MLE of the MNL model for the current bootstrap sample.
    new_id_col : str.
        Denotes the new column that specifies the bootstrap observation ids for
        choice model estimation.
    num_params : non-negative int.
        The number of parameters in the MLE of the `orig_model_obj`.
    mnl_spec : OrderedDict or None.
        If `orig_model_obj` is not a MNL model, then `mnl_spec` should be an
        OrderedDict that contains the specification dictionary used to estimate
        the MNL model that will provide starting values for the final estimated
        model. If `orig_model_obj` is a MNL model, then `mnl_spec` may be None.
    mnl_names : OrderedDict or None.
        If `orig_model_obj` is not a MNL model, then `mnl_names` should be an
        OrderedDict that contains the name dictionary used to initialize the
        MNL model that will provide starting values for the final estimated
        model. If `orig_model_obj` is a MNL, then `mnl_names` may be None.
    mnl_init_vals : 1D ndarray or None.
        If `orig_model_obj` is not a MNL model, then `mnl_init_vals` should be
        a 1D ndarray. `mnl_init_vals` should denote the initial values used to
        estimate the MNL model that provides starting values for the final
        desired model. If `orig_model_obj` is a MNL model, then `mnl_init_vals`
        may be None.
    mnl_fit_kwargs : dict or None.
        If `orig_model_obj` is not a MNL model, then `mnl_fit_kwargs` should be
        a dict. `mnl_fit_kwargs` should denote the keyword arguments used when
        calling the `fit_mle` function of the MNL model that will provide
        starting values to the desired choice model. If `orig_model_obj` is a
        MNL model, then `mnl_fit_kwargs` may be None.
    extract_init_vals : callable or None, optional.
        Should accept 3 arguments, in the following order. First, it should
        accept `orig_model_obj`. Second, it should accept a pandas Series of
        the estimated parameters from the MNL model. The index of the Series
        will be the names of the coefficients from `mnl_names`. Thirdly, it
        should accept an int denoting the number of parameters in the desired
        choice model. The callable should return a 1D ndarray of starting
        values for the desired choice model. Default == None.
    fit_kwargs : dict.
        Denotes the keyword arguments to be used when estimating the desired
        choice model using the current bootstrap sample (`new_df`). All such
        kwargs will be directly passed to the `fit_mle` method of the desired
        model object.

    Returns
    -------
    final_point : dict.
        The dictionary returned by `scipy.optimize` after estimating the
        desired choice model.
    """
    # Get the MNL point estimate for the parameters of this bootstrap sample.
    mnl_point, mnl_obj = get_mnl_point_est(orig_model_obj,
                                           new_df,
                                           new_id_col,
                                           num_params,
                                           mnl_spec,
                                           mnl_names,
                                           mnl_init_vals,
                                           mnl_fit_kwargs)
    mnl_point_series = pd.Series(mnl_point["x"], index=mnl_obj.ind_var_names)

    # Denote the MNL point estimate as our final point estimate if the final
    # model we're interested in is an MNL.
    if orig_model_obj.model_type == model_type_to_display_name["MNL"]:
        final_point = mnl_point
    else:
        # Determine the function to be used when extracting the initial values
        # for the final model from the MNL MLE point estimate.
        if extract_init_vals is None:
            extraction_func = extract_default_init_vals
        else:
            extraction_func = extract_init_vals

        # Extract the initial values
        default_init_vals =\
            extraction_func(orig_model_obj, mnl_point_series, num_params)

        # Get the keyword arguments needed to initialize the new model object.
        model_kwargs = get_model_creation_kwargs(orig_model_obj)

        # Create a new model object
        new_obj =\
            pl.create_choice_model(data=new_df,
                                   alt_id_col=orig_model_obj.alt_id_col,
                                   obs_id_col=new_id_col,
                                   choice_col=orig_model_obj.choice_col,
                                   specification=orig_model_obj.specification,
                                   **model_kwargs)

        # Be sure to add 'just_point' to perform pure point estimation.
        if 'just_point' not in fit_kwargs:
            fit_kwargs['just_point'] = True

        # Fit the model with new data, and return the point estimate dict.
        final_point = new_obj.fit_mle(default_init_vals, **fit_kwargs)

    return final_point



"""
@author:    Timothy Brathwaite
@name:      Bootstrap Calculations
@summary:   This module provides functions to calculate the bootstrap
            confidence intervals using the 'percentile' and
            'bias-corrected and accelerated' methods.
"""
import numpy as np
from scipy.stats import norm

from .bootstrap_utils import check_conf_percentage_validity
from .bootstrap_utils import ensure_samples_is_ndim_ndarray
from .bootstrap_utils import get_alpha_from_conf_percentage
from .bootstrap_utils import combine_conf_endpoints

# Create a value to be used to avoid numeric underflow.
MIN_COMP_VALUE = 1e-16


def calc_percentile_interval(bootstrap_replicates, conf_percentage):
    """
    Calculate bootstrap confidence intervals based on raw percentiles of the
    bootstrap distribution of samples.

    Parameters
    ----------
    bootstrap_replicates : 2D ndarray.
        Each row should correspond to a different bootstrap parameter sample.
        Each column should correspond to an element of the parameter vector
        being estimated.
    conf_percentage : scalar in the interval (0.0, 100.0).
        Denotes the confidence-level of the returned confidence interval. For
        instance, to calculate a 95% confidence interval, pass `95`.

    Returns
    -------
    conf_intervals : 2D ndarray.
        The shape of the returned array will be `(2, samples.shape[1])`. The
        first row will correspond to the lower value in the confidence
        interval. The second row will correspond to the upper value in the
        confidence interval. There will be one column for each element of the
        parameter vector being estimated.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 12.5 and Section 13.3. See Equation 13.3.

    Notes
    -----
    This function differs slightly from the actual percentile bootstrap
    procedure described in Efron and Tibshirani (1994). To ensure that the
    returned endpoints of one's bootstrap confidence intervals are actual
    values that were observed in the bootstrap distribution, both the procedure
    of Efron and Tibshirani and this function make more conservative confidence
    intervals. However, this function uses a simpler (and in some cases less
    conservative) correction than that of Efron and Tibshirani.
    """
    # Check validity of arguments
    check_conf_percentage_validity(conf_percentage)
    ensure_samples_is_ndim_ndarray(bootstrap_replicates, ndim=2)
    # Get the alpha * 100% value
    alpha = get_alpha_from_conf_percentage(conf_percentage)
    # Get the lower and upper percentiles that demarcate the desired interval.
    lower_percent = alpha / 2.0
    upper_percent = 100.0 - lower_percent
    # Calculate the lower and upper endpoints of the confidence intervals.
    # Note that the particular choices of interpolation methods are made in
    # order to produce conservatively wide confidence intervals and ensure that
    # all returned endpoints in the confidence intervals are actually observed
    # in the bootstrap distribution. This is in accordance with the spirit of
    # Efron and Tibshirani (1994).
    lower_endpoint = np.percentile(bootstrap_replicates,
                                   lower_percent,
                                   interpolation='lower',
                                   axis=0)
    upper_endpoint = np.percentile(bootstrap_replicates,
                                   upper_percent,
                                   interpolation='higher',
                                   axis=0)
    # Combine the enpoints into a single ndarray.
    conf_intervals = combine_conf_endpoints(lower_endpoint, upper_endpoint)
    return conf_intervals


def calc_bias_correction_bca(bootstrap_replicates, mle_estimate):
    """
    Calculate the bias correction for the Bias Corrected and Accelerated (BCa)
    bootstrap confidence intervals.

    Parameters
    ----------
    bootstrap_replicates : 2D ndarray.
        Each row should correspond to a different bootstrap parameter sample.
        Each column should correspond to an element of the parameter vector
        being estimated.
    mle_estimate : 1D ndarray.
        The original dataset's maximum likelihood point estimate. Should have
        one elements for each component of the estimated parameter vector.

    Returns
    -------
    bias_correction : 1D ndarray.
        There will be one element for each element in `mle_estimate`. Elements
        denote the bias correction factors for each component of the parameter
        vector.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 14.3, Equation 14.14.
    """
    numerator = (bootstrap_replicates < mle_estimate[None, :]).sum(axis=0)
    denominator = float(bootstrap_replicates.shape[0])
    bias_correction = norm.ppf(numerator / denominator)
    return bias_correction


def calc_acceleration_bca(jackknife_replicates):
    """
    Calculate the acceleration constant for the Bias Corrected and Accelerated
    (BCa) bootstrap confidence intervals.

    Parameters
    ----------
    jackknife_replicates : 2D ndarray.
        Each row should correspond to a different jackknife parameter sample,
        formed by deleting a particular observation and then re-estimating the
        desired model. Each column should correspond to an element of the
        parameter vector being estimated.

    Returns
    -------
    acceleration : 1D ndarray.
        There will be one element for each element in `mle_estimate`. Elements
        denote the acceleration factors for each component of the parameter
        vector.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 14.3, Equation 14.15.
    """
    # Get the mean of the bootstrapped statistics.
    jackknife_mean = jackknife_replicates.mean(axis=0)[None, :]
    # Calculate the differences between the mean of the bootstrapped statistics
    differences = jackknife_mean - jackknife_replicates
    numerator = (differences**3).sum(axis=0)
    denominator = 6 * ((differences**2).sum(axis=0))**1.5
    # guard against division by zero. Note that this guard shouldn't distort
    # the computational results since the numerator should be zero whenever the
    # denominator is zero.
    zero_denom = np.where(denominator == 0)
    denominator[zero_denom] = MIN_COMP_VALUE
    # Compute the acceleration.
    acceleration = numerator / denominator
    return acceleration


def calc_lower_bca_percentile(alpha_percent, bias_correction, acceleration):
    """
    Calculate the lower values of the Bias Corrected and Accelerated (BCa)
    bootstrap confidence intervals.

    Parameters
    ----------
    alpha_percent : float in (0.0, 100.0).
        `100 - confidence_percentage`, where `confidence_percentage` is the
        confidence level (such as 95%), expressed as a percent.
    bias_correction : 1D ndarray.
        There will be one element for each element in `mle_estimate`. Elements
        denote the bias correction factors for each component of the parameter
        vector.
    acceleration : 1D ndarray.
        There will be one element for each element in `mle_estimate`. Elements
        denote the acceleration factors for each component of the parameter
        vector.

    Returns
    -------
    lower_percentile : 1D ndarray.
        There will be one element for each element in `mle_estimate`. Elements
        denote the smaller values in the confidence interval for each component
        of the parameter vector.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 14.3, Equation 14.10.

    Notes
    -----
    The `alpha` used in this function is different from the `alpha` used in
    Efron and Tibshirani (1994). The `alpha` used in this function must be
    converted to a decimal (by dividing by 100) and then divided by 2 (to
    account for the equal-tailed nature of the confidence interval) in order to
    be made equivalent to the `alpha` in Efron and Tibshirani (1994).
    """
    z_lower = norm.ppf(alpha_percent / (100.0 * 2))
    numerator = bias_correction + z_lower
    denominator = 1 - acceleration * numerator
    lower_percentile =\
        norm.cdf(bias_correction + numerator / denominator) * 100
    return lower_percentile


def calc_upper_bca_percentile(alpha_percent, bias_correction, acceleration):
    """
    Calculate the lower values of the Bias Corrected and Accelerated (BCa)
    bootstrap confidence intervals.

    Parameters
    ----------
    alpha_percent : float in (0.0, 100.0).
        `100 - confidence_percentage`, where `confidence_percentage` is the
        confidence level (such as 95%), expressed as a percent.
    bias_correction : 1D ndarray.
        There will be one element for each element in `mle_estimate`. Elements
        denote the bias correction factors for each component of the parameter
        vector.
    acceleration : 1D ndarray.
        There will be one element for each element in `mle_estimate`. Elements
        denote the acceleration factors for each component of the parameter
        vector.

    Returns
    -------
    upper_percentile : 1D ndarray.
        There will be one element for each element in `mle_estimate`. Elements
        denote the larger values in the confidence interval for each component
        of the parameter vector.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 14.3, Equation 14.10.

    Notes
    -----
    The `alpha` used in this function is different from the `alpha` used in
    Efron and Tibshirani (1994). The `alpha` used in this function must be
    converted to a decimal (by dividing by 100) and then divided by 2 (to
    account for the equal-tailed nature of the confidence interval) in order to
    be made equivalent to the `alpha` in Efron and Tibshirani (1994).
    """
    z_upper = norm.ppf(1 - alpha_percent / (100.0 * 2))
    numerator = bias_correction + z_upper
    denominator = 1 - acceleration * numerator
    upper_percentile =\
        norm.cdf(bias_correction + numerator / denominator) * 100
    return upper_percentile


def calc_bca_interval(bootstrap_replicates,
                      jackknife_replicates,
                      mle_params,
                      conf_percentage):
    """
    Calculate 'bias-corrected and accelerated' bootstrap confidence intervals.

    Parameters
    ----------
    bootstrap_replicates : 2D ndarray.
        Each row should correspond to a different bootstrap parameter sample.
        Each column should correspond to an element of the parameter vector
        being estimated.
    jackknife_replicates : 2D ndarray.
        Each row should correspond to a different jackknife parameter sample,
        formed by deleting a particular observation and then re-estimating the
        desired model. Each column should correspond to an element of the
        parameter vector being estimated.
    mle_params : 1D ndarray.
        The original dataset's maximum likelihood point estimate. Should have
        the same number of elements as `samples.shape[1]`.
    conf_percentage : scalar in the interval (0.0, 100.0).
        Denotes the confidence-level of the returned confidence interval. For
        instance, to calculate a 95% confidence interval, pass `95`.

    Returns
    -------
    conf_intervals : 2D ndarray.
        The shape of the returned array will be `(2, samples.shape[1])`. The
        first row will correspond to the lower value in the confidence
        interval. The second row will correspond to the upper value in the
        confidence interval. There will be one column for each element of the
        parameter vector being estimated.

    References
    ----------
    Efron, Bradley, and Robert J. Tibshirani. An Introduction to the Bootstrap.
        CRC press, 1994. Section 14.3.
    DiCiccio, Thomas J., and Bradley Efron. "Bootstrap confidence intervals."
        Statistical science (1996): 189-212.
    """
    # Check validity of arguments
    check_conf_percentage_validity(conf_percentage)
    ensure_samples_is_ndim_ndarray(bootstrap_replicates, ndim=2)
    ensure_samples_is_ndim_ndarray(jackknife_replicates,
                                   name='jackknife', ndim=2)
    # Calculate the alpha * 100% value
    alpha_percent = get_alpha_from_conf_percentage(conf_percentage)
    # Estimate the bias correction for the bootstrap samples
    bias_correction =\
        calc_bias_correction_bca(bootstrap_replicates, mle_params)
    # Estimate the acceleration
    acceleration = calc_acceleration_bca(jackknife_replicates)
    # Get the lower and upper percent value for the raw bootstrap samples.
    lower_percents =\
        calc_lower_bca_percentile(alpha_percent, bias_correction, acceleration)
    upper_percents =\
        calc_upper_bca_percentile(alpha_percent, bias_correction, acceleration)
    # Get the lower and upper endpoints for the desired confidence intervals.
    lower_endpoints = np.diag(np.percentile(bootstrap_replicates,
                                            lower_percents,
                                            interpolation='lower',
                                            axis=0))
    upper_endpoints = np.diag(np.percentile(bootstrap_replicates,
                                            upper_percents,
                                            interpolation='higher',
                                            axis=0))
    # Combine the enpoints into a single ndarray.
    conf_intervals = combine_conf_endpoints(lower_endpoints, upper_endpoints)
    return conf_intervals



"""
@author:    Timothy Brathwaite
@name:      Bootstrap Utilities
@summary:   This module provides helpful functions for calculating the
            bootstrap confidence intervals.
"""
from numbers import Number
import numpy as np


def check_conf_percentage_validity(conf_percentage):
    """
    Ensures that `conf_percentage` is in (0, 100). Raises a helpful ValueError
    if otherwise.
    """
    msg = "conf_percentage MUST be a number between 0.0 and 100."
    condition_1 = isinstance(conf_percentage, Number)
    if not condition_1:
        raise ValueError(msg)
    else:
        condition_2 = 0 < conf_percentage < 100
        if not condition_2:
            raise ValueError(msg)
    return None


def ensure_samples_is_ndim_ndarray(samples, name='bootstrap', ndim=2):
    """
    Ensures that `samples` is an `ndim` numpy array. Raises a helpful
    ValueError if otherwise.
    """
    assert isinstance(ndim, int)
    assert isinstance(name, str)
    if not isinstance(samples, np.ndarray) or not (samples.ndim == ndim):
        sample_name = name + "_samples"
        msg = "`{}` MUST be a {}D ndarray.".format(sample_name, ndim)
        raise ValueError(msg)
    return None


def get_alpha_from_conf_percentage(conf_percentage):
    """
    Calculates `100 - conf_percentage`, which is useful for calculating alpha
    levels.
    """
    return 100.0 - conf_percentage


def combine_conf_endpoints(lower_array, upper_array):
    """
    Concatenates upper and lower endpoint arrays for a given confidence level.
    """
    return np.concatenate([lower_array[None, :], upper_array[None, :]], axis=0)



"""
@name:      Estimator Constructor
@author:    Timothy Brathwaite
@summary:   Contains functions necessary for constructing the Estimation
            Objects used to provide convenience functions when estimating
            PyLogit's various choice models.
"""
from __future__ import absolute_import

import numpy as np

from .display_names import model_type_to_display_name as display_name_dict

from .mixed_logit import MixedEstimator
from .mixed_logit import split_param_vec as mixed_split_params

from .nested_logit import NestedEstimator
from .nested_logit import split_param_vec as nested_split_params

from .conditional_logit import MNLEstimator
from .conditional_logit import split_param_vec as mnl_split_params

from .clog_log import ClogEstimator
from .clog_log import split_param_vec as clog_split_params

from .asym_logit import AsymEstimator
from .asym_logit import split_param_vec as asym_split_params

from .scobit import ScobitEstimator
from .scobit import split_param_vec as scobit_split_params

from .uneven_logit import UnevenEstimator
from .uneven_logit import split_param_vec as uneven_split_params

# Map the displayed model types to the internal model names.
display_name_to_model_type = {v: k for k, v in display_name_dict.items()}

# Map the internal model types to their appropriate estimator and split params
# functions
model_type_to_resources =\
    {"MNL": {'estimator': MNLEstimator, 'split_func': mnl_split_params},
     "Asym": {'estimator': AsymEstimator, 'split_func': asym_split_params},
     "Cloglog": {'estimator': ClogEstimator, 'split_func': clog_split_params},
     "Scobit": {'estimator': ScobitEstimator,
                'split_func': scobit_split_params},
     "Uneven": {'estimator': UnevenEstimator,
                'split_func': uneven_split_params},
     "Nested Logit": {'estimator': NestedEstimator,
                      'split_func': nested_split_params},
     "Mixed Logit": {'estimator': MixedEstimator,
                     'split_func': mixed_split_params}}


def create_estimation_obj(model_obj,
                          init_vals,
                          mappings=None,
                          ridge=None,
                          constrained_pos=None,
                          weights=None):
    """
    Should return a model estimation object corresponding to the model type of
    the `model_obj`.

    Parameters
    ----------
    model_obj : an instance or sublcass of the MNDC class.
    init_vals : 1D ndarray.
        The initial values to start the estimation process with. In the
        following order, there should be one value for each nest coefficient,
        shape parameter, outside intercept parameter, or index coefficient that
        is being estimated.
    mappings : OrderedDict or None, optional.
        Keys will be `["rows_to_obs", "rows_to_alts", "chosen_row_to_obs",
        "rows_to_nests"]`. The value for `rows_to_obs` will map the rows of
        the `long_form` to the unique observations (on the columns) in
        their order of appearance. The value for `rows_to_alts` will map
        the rows of the `long_form` to the unique alternatives which are
        possible in the dataset (on the columns), in sorted order--not
        order of appearance. The value for `chosen_row_to_obs`, if not
        None, will map the rows of the `long_form` that contain the chosen
        alternatives to the specific observations those rows are associated
        with (denoted by the columns). The value of `rows_to_nests`, if not
        None, will map the rows of the `long_form` to the nest (denoted by
        the column) that contains the row's alternative. Default == None.
    ridge : int, float, long, or None, optional.
        Determines whether or not ridge regression is performed. If a
        scalar is passed, then that scalar determines the ridge penalty for
        the optimization. The scalar should be greater than or equal to
        zero. Default `== None`.
    constrained_pos : list or None, optional.
        Denotes the positions of the array of estimated parameters that are
        not to change from their initial values. If a list is passed, the
        elements are to be integers where no such integer is greater than
        `init_vals.size.` Default == None.
    weights : 1D ndarray.
        Should contain the weights for each corresponding observation for each
        row of the long format data.
    """
    # Get the mapping matrices for each model
    mapping_matrices =\
        model_obj.get_mappings_for_fit() if mappings is None else mappings
    # Create the zero vector for each model.
    zero_vector = np.zeros(init_vals.shape[0])
    # Get the internal model name
    internal_model_name = display_name_to_model_type[model_obj.model_type]
    # Get the split parameter function and estimator class for this model.
    estimator_class, current_split_func =\
        (model_type_to_resources[internal_model_name]['estimator'],
         model_type_to_resources[internal_model_name]['split_func'])
    # Create the estimator instance that is desired.
    estimation_obj = estimator_class(model_obj,
                                     mapping_matrices,
                                     ridge,
                                     zero_vector,
                                     current_split_func,
                                     constrained_pos,
                                     weights=weights)
    # Return the created object
    return estimation_obj



# -*- coding: utf-8 -*-
"""
Created on Sun Jul  3 11:26:33 2016

@module:    nested_logit.py
@name:      Nested Logit Model
@author:    Timothy Brathwaite
@summary:   Contains functions necessary for estimating nested logit models
            (with the help of the "base_multinomial_cm.py" file).
"""
from __future__ import absolute_import

import warnings
import numpy as np

from . import nested_choice_calcs as nc
from . import base_multinomial_cm_v2 as base_mcm
from .display_names import model_type_to_display_name
from .estimation import EstimationObj
from .estimation import estimate

# Alias necessary functions from the base multinomial choice model module
general_log_likelihood = nc.calc_nested_log_likelihood
general_gradient = nc.calc_nested_gradient
general_calc_probabilities = nc.calc_nested_probs
bhhh_approx = nc.calc_bhhh_hessian_approximation

# Create a warning string that will be issued if ridge regression is performed.
_msg_3 = "NOTE: An L2-penalized regression is being performed. The "
_msg_4 = "reported standard errors and robust standard errors "
_msg_5 = "***WILL BE INCORRECT***."
_ridge_warning_msg = _msg_3 + _msg_4 + _msg_5


# Create a function that will identify degenerate nests
def identify_degenerate_nests(nest_spec):
    """
    Identify the nests within nest_spec that are degenerate, i.e. those nests
    with only a single alternative within the nest.

    Parameters
    ----------
    nest_spec : OrderedDict.
        Keys are strings that define the name of the nests. Values are lists
        of alternative ids, denoting which alternatives belong to which nests.
        Each alternative id must only be associated with a single nest!

    Returns
    -------
    list.
        Will contain the positions in the list of keys from `nest_spec` that
        are degenerate.
    """
    degenerate_positions = []
    for pos, key in enumerate(nest_spec):
        if len(nest_spec[key]) == 1:
            degenerate_positions.append(pos)
    return degenerate_positions


# Define a functionto split the combined parameter array into nest coefficients
# and index coefficients.
def split_param_vec(all_params, rows_to_nests, return_all_types=False):
    """
    Parameters
    ----------
    all_params : 1D ndarray.
        Should contain all of the parameters being estimated (i.e. all the
        nest coefficients and all of the index coefficients). All elements
        should be ints, floats, or longs.
    rows_to_nests : 2D scipy sparse array.
        There should be one row per observation per available alternative and
        one column per nest. This matrix maps the rows of the design matrix to
        the unique nests (on the columns).
    return_all_types : bool, optional.
        Determines whether or not a tuple of 4 elements will be returned (with
        one element for the nest, shape, intercept, and index parameters for
        this model). If False, a tuple of 2 elements will be returned, as
        described below. The tuple will contain the nest parameters and the
        index coefficients.

    Returns
    -------
    orig_nest_coefs : 1D ndarray.
        The nest coefficients being used for estimation. Note that these values
        are the logit of the inverse of the scale parameters for each lower
        level nest.
    index_coefs : 1D ndarray.
        The coefficients of the index being used for this nested logit model.

    Note
    ----
    If `return_all_types == True` then the function will return a tuple of four
    objects. In order, these objects will either be None or the arrays
    representing the arrays corresponding to the nest, shape, intercept, and
    index parameters.
    """
    # Split the array of all coefficients
    num_nests = rows_to_nests.shape[1]
    orig_nest_coefs = all_params[:num_nests]
    index_coefs = all_params[num_nests:]

    if return_all_types:
        return orig_nest_coefs, None, None, index_coefs
    else:
        return orig_nest_coefs, index_coefs


class NestedEstimator(EstimationObj):
    """
    Estimation object for the 2-level Nested Logit Model.

    Parameters
    ----------
    model_obj : a pylogit.base_multinomial_cm_v2.MNDC_Model instance.
        Should contain the following attributes:

          - alt_IDs
          - choices
          - design
          - intercept_ref_position
          - shape_ref_position
          - utility_transform
    mapping_res : dict.
        Should contain the scipy sparse matrices that map the rows of the long
        format dataframe to various other objects such as the available
        alternatives, the unique observations, etc. The keys that it must have
        are `['rows_to_obs', 'rows_to_alts', 'chosen_row_to_obs']`
    ridge : int, float, long, or None.
            Determines whether or not ridge regression is performed. If a
            scalar is passed, then that scalar determines the ridge penalty for
            the optimization. The scalar should be greater than or equal to
            zero.
    zero_vector : 1D ndarray.
        Determines what is viewed as a "null" set of parameters. It is
        explicitly passed because some parameters (e.g. parameters that must be
        greater than zero) have their null values at values other than zero.
    split_params : callable.
        Should take a vector of parameters, `mapping_res['rows_to_alts']`, and
        model_obj.design as arguments. Should return a tuple containing
        separate arrays for the model's shape, outside intercept, and index
        coefficients. For each of these arrays, if this model does not contain
        the particular type of parameter, the callable should place a `None` in
        its place in the tuple.
    constrained_pos : list or None, optional.
        Denotes the positions of the array of estimated parameters that are
        not to change from their initial values. If a list is passed, the
        elements are to be integers where no such integer is greater than
        `init_values.size.` Default == None.
    weights : 1D ndarray or None, optional.
        Allows for the calculation of weighted log-likelihoods. The weights can
        represent various things. In stratified samples, the weights may be
        the proportion of the observations in a given strata for a sample in
        relation to the proportion of observations in that strata in the
        population. In latent class models, the weights may be the probability
        of being a particular class.

    """
    def __init__(self,
                 model_obj,
                 mapping_dict,
                 ridge,
                 zero_vector,
                 split_params,
                 constrained_pos=None,
                 weights=None):
        super(NestedEstimator, self).__init__(model_obj,
                                              mapping_dict,
                                              ridge,
                                              zero_vector,
                                              split_params,
                                              constrained_pos=constrained_pos,
                                              weights=weights)

        return None

    def check_length_of_initial_values(self, init_values):
        """
        Ensures that the initial values are of the correct length.
        """
        # Figure out how many shape parameters we should have and how many
        # index coefficients we should have
        num_nests = self.rows_to_nests.shape[1]
        num_index_coefs = self.design.shape[1]

        assumed_param_dimensions = num_index_coefs + num_nests
        if init_values.shape[0] != assumed_param_dimensions:
            msg = "The initial values are of the wrong dimension"
            msg_1 = "It should be of dimension {}"
            msg_2 = "But instead it has dimension {}"
            raise ValueError(msg +
                             msg_1.format(assumed_param_dimensions) +
                             msg_2.format(init_values.shape[0]))

        return None

    def convenience_split_params(self, params, return_all_types=False):
        """
        Splits parameter vector into nest parameters and index parameters.

        Parameters
        ----------
        all_params : 1D ndarray.
            Should contain all of the parameters being estimated (i.e. all the
            nest coefficients and all of the index coefficients). All elements
            should be ints, floats, or longs.
        rows_to_nests : 2D scipy sparse array.
            There should be one row per observation per available alternative
            and one column per nest. This matrix maps the rows of the design
            matrix to the unique nests (on the columns).
        return_all_types : bool, optional.
            Determines whether or not a tuple of 4 elements will be returned
            (with one element for the nest, shape, intercept, and index
            parameters for this model). If False, a tuple of 2 elements will
            be returned, as described below. The tuple will contain the nest
            parameters and the index coefficients.

        Returns
        -------
        orig_nest_coefs : 1D ndarray.
            The nest coefficients being used for estimation. Note that these
            values are the logit of the inverse of the scale parameters for
            each lower level nest.
        index_coefs : 1D ndarray.
            The index coefficients of this nested logit model.

        Note
        ----
        If `return_all_types == True` then the function will return a tuple of
        four objects. In order, these objects will either be None or the arrays
        representing the arrays corresponding to the nest, shape, intercept,
        and index parameters.
        """
        return split_param_vec(params,
                               self.rows_to_nests,
                               return_all_types=return_all_types)

    def convenience_calc_probs(self, params):
        """
        Calculates the probabilities of the chosen alternative, and the long
        format probabilities for this model and dataset.
        """
        orig_nest_coefs, betas = self.convenience_split_params(params)
        natural_nest_coefs = nc.naturalize_nest_coefs(orig_nest_coefs)

        args = [natural_nest_coefs,
                betas,
                self.design,
                self.rows_to_obs,
                self.rows_to_nests]
        kwargs = {"chosen_row_to_obs": self.chosen_row_to_obs,
                  "return_type": "long_and_chosen_probs"}

        probability_results = general_calc_probabilities(*args, **kwargs)

        return probability_results

    def convenience_calc_log_likelihood(self, params):
        """
        Calculates the log-likelihood for this model and dataset.
        """
        orig_nest_coefs, betas = self.convenience_split_params(params)
        natural_nest_coefs = nc.naturalize_nest_coefs(orig_nest_coefs)

        args = [natural_nest_coefs,
                betas,
                self.design,
                self.rows_to_obs,
                self.rows_to_nests,
                self.choice_vector]
        kwargs = {"ridge": self.ridge, "weights": self.weights}

        log_likelihood = general_log_likelihood(*args, **kwargs)

        return log_likelihood

    def convenience_calc_gradient(self, params):
        """
        Calculates the gradient of the log-likelihood for this model / dataset.
        """
        orig_nest_coefs, betas = self.convenience_split_params(params)

        args = [orig_nest_coefs,
                betas,
                self.design,
                self.choice_vector,
                self.rows_to_obs,
                self.rows_to_nests]

        return general_gradient(*args, ridge=self.ridge, weights=self.weights)

    def convenience_calc_hessian(self, params):
        """
        Calculates the hessian of the log-likelihood for this model / dataset.
        Note that this function name is INCORRECT with regard to the actual
        actions performed. The Nested Logit model uses the BHHH approximation
        to the Fisher Information Matrix in place of the actual hessian.
        """
        orig_nest_coefs, betas = self.convenience_split_params(params)

        args = [orig_nest_coefs,
                betas,
                self.design,
                self.choice_vector,
                self.rows_to_obs,
                self.rows_to_nests]

        approx_hess =\
            bhhh_approx(*args, ridge=self.ridge, weights=self.weights)

        # Account for the contrained position when presenting the results of
        # the approximate hessian.
        if self.constrained_pos is not None:
            for idx_val in self.constrained_pos:
                approx_hess[idx_val, :] = 0
                approx_hess[:, idx_val] = 0
                approx_hess[idx_val, idx_val] = -1

        return approx_hess

    def convenience_calc_fisher_approx(self, params):
        """
        Calculates the BHHH approximation of the Fisher Information Matrix for
        this model / dataset. Note that this function name is INCORRECT with
        regard to the actual actions performed. The Nested Logit model uses a
        placeholder for the BHHH approximation of the Fisher Information Matrix
        because the BHHH approximation is already being used to approximate the
        hessian.

        This placeholder allows calculation of a value for the 'robust'
        standard errors, even though such a value is not useful since it is not
        correct...
        """
        placeholder_bhhh = np.diag(-1 * np.ones(params.shape[0]))

        return placeholder_bhhh


# Define the nested logit model class
class NestedLogit(base_mcm.MNDC_Model):
    """
    Parameters
    ----------
    data : string or pandas dataframe.
        If string, data should be an absolute or relative path to a CSV file
        containing the long format data for this choice model. Note long format
        is has one row per available alternative for each observation. If
        pandas dataframe, the dataframe should be the long format data for the
        choice model.
    alt_id_col : string.
        Should denote the column in `data` that contains the identifiers of
        each row's corresponding alternative.
    obs_id_col : string.
        Should denote the column in `data` that contains the identifiers of
        each row's corresponding observation.
    choice_col : string.
        Should denote the column in data which contains the ones and zeros that
        denote whether or not the given row corresponds to the chosen
        alternative for the given individual.
    specification : OrderedDict.
        Keys are a proper subset of the columns in `long_form_df`. Values are
        either a list or a single string, `'all_diff' or 'all_same'`. If a
        list, the elements should be:

            - single objects that are within the alternative ID column of
              `long_form_df`
            - lists of objects that are within the alternative ID column of
              `long_form_df`.

        For each single object in the list, a unique column will be created
        (i.e. there will be a unique coefficient for that variable in the
        corresponding utility equation of the corresponding alternative). For
        lists within the specification_dict values, a single column will be
        created for all the alternatives within iterable (i.e. there will be
        one common coefficient for the variables in the iterable).
    names : OrderedDict, optional.
        Should have the same keys as `specification`. For each key:

            - if the corresponding value in `specification` is `'all_same'`,
              then there should be a single string as the value in names.
            - if the corresponding value in `specification` is `'all_diff'`,
              then there should be a list of strings as the value in names.
              There should be one string in the value in names for each
              possible alternative.
            - if the corresponding value in `specification` is a list, then
              there should be a list of strings as the value in names. There
              should be one string the value in names per item in the value in
              `specification`.

        Default == None.
    nest_spec : OrderedDict, optional.
        This kwarg MUST be passed for nested logit models. It is optional only
        for api compatibility. Keys are strings that define the name of the
        nests. Values are lists of alternative ids, denoting which alternatives
        belong to which nests. Each alternative id must only be associated with
        a single nest! Default `== None`.
    model_type : str, optional.
        Denotes the model type of the choice model being instantiated.
        Default == "".
    """
    def __init__(self,
                 data,
                 alt_id_col,
                 obs_id_col,
                 choice_col,
                 specification,
                 names=None,
                 nest_spec=None,
                 **kwargs):
        ##########
        # Print a helpful message if nest_spec has not been included
        ##########
        condition_1 = nest_spec is None and "nest_spec" not in kwargs
        condition_2 = "nest_spec" in kwargs and kwargs["nest_spec"] is None
        missing_nest_spec = condition_1 or condition_2
        if missing_nest_spec:
            msg = "The Nested Logit Model REQUIRES a nest specification dict."
            raise ValueError(msg)

        ##########
        # Carry out the common instantiation process for all choice models
        ##########
        model_name = model_type_to_display_name["Nested Logit"]
        super(NestedLogit, self).__init__(data,
                                          alt_id_col,
                                          obs_id_col,
                                          choice_col,
                                          specification,
                                          names=names,
                                          nest_spec=nest_spec,
                                          model_type=model_name)

        ##########
        # Store the utility transform function
        ##########
        self.utility_transform = lambda x, *args, **kwargs: x[:, None]

        return None

    def fit_mle(self,
                init_vals,
                constrained_pos=None,
                print_res=True,
                method="BFGS",
                loss_tol=1e-06,
                gradient_tol=1e-06,
                maxiter=1000,
                ridge=None,
                just_point=False,
                **kwargs):
        """
        Parameters
        ----------
        init_vals : 1D ndarray.
            Should containn the initial values to start the optimization
            process with. There should be one value for each nest parameter
            and utility coefficient. Nest parameters not being estimated
            should still be included. Handle these parameters using the
            `constrained_pos` kwarg.
        constrained_pos : list, or None, optional.
            Denotes the positions of the array of estimated parameters that are
            not to change from their initial values. If a list is passed, the
            elements are to be integers where no such integer is greater than
            `init_values.size`.
        print_res : bool, optional.
            Determines whether the timing and initial and final log likelihood
            results will be printed as they they are determined.
        method : str, optional.
            Should be a valid string which can be passed to
            `scipy.optimize.minimize`. Determines the optimization algorithm
            which is used for this problem.
        loss_tol : float, optional.
            Determines the tolerance on the difference in objective function
            values from one iteration to the next which is needed to determine
            convergence. Default `== 1e-06`.
        gradient_tol : float, optional.
            Determines the tolerance on the difference in gradient values from
            one iteration to the next which is needed to determine convergence.
            Default `== 1e-06`.
        ridge : int, float, long, or None, optional.
            Determines whether ridge regression is performed. If a scalar is
            passed, then that scalar determines the ridge penalty for the
            optimization. Default `== None`.
        just_point : bool, optional.
            Determines whether (True) or not (False) calculations that are non-
            critical for obtaining the maximum likelihood point estimate will
            be performed. If True, this function will return the results
            dictionary from scipy.optimize. Default == False.

        Returns
        -------
        None. Estimation results are saved to the model instance.
        """
        # Check integrity of passed arguments
        kwargs_to_be_ignored = ["init_shapes", "init_intercepts", "init_coefs"]
        if any([x in kwargs for x in kwargs_to_be_ignored]):
            msg = "Nested Logit model does not use the following kwargs:\n{}"
            msg_2 = "Remove such kwargs and pass a single init_vals argument"
            raise ValueError(msg.format(kwargs_to_be_ignored) + msg_2)

        # Store the optimization method
        self.optimization_method = method

        # Store the ridge parameter
        self.ridge_param = ridge

        if ridge is not None:
            warnings.warn(_ridge_warning_msg)

        # Construct the mappings from alternatives to observations and from
        # chosen alternatives to observations
        mapping_res = self.get_mappings_for_fit()

        # Determine the degenerate nests whose nesting parameters are to be
        # constrained to one. Note the following functions assume that the nest
        # parameters are placed before the index coefficients.
        fixed_params = identify_degenerate_nests(self.nest_spec)

        # Include the user specified parameters that are to be constrained to
        # their initial values
        if constrained_pos is not None:
            fixed_params.extend(constrained_pos)
        final_constrained_pos = sorted(list(set(fixed_params)))

        # Create the estimation object
        zero_vector = np.zeros(init_vals.shape)
        estimator_args = [self,
                          mapping_res,
                          ridge,
                          zero_vector,
                          split_param_vec]
        estimator_kwargs = {"constrained_pos": final_constrained_pos}
        nested_estimator = NestedEstimator(*estimator_args,
                                           **estimator_kwargs)

        # Perform one final check on the length of the initial values
        nested_estimator.check_length_of_initial_values(init_vals)

        # Get the estimation results
        estimation_res = estimate(init_vals,
                                  nested_estimator,
                                  method,
                                  loss_tol,
                                  gradient_tol,
                                  maxiter,
                                  print_res,
                                  use_hessian=True,
                                  just_point=just_point)

        if not just_point:
            # Store the estimation results
            self.store_fit_results(estimation_res)

            return None
        else:
            return estimation_res
